// -*- c-basic-offset: 4 -*-

/**
 *  @author Pablo d'Angelo <pablo.dangelo@web.de>
 *
 *  $Id: panorama.cpp,v 1.1.1.1 2003/04/06 18:39:00 dangelo Exp $
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <sstream>

#include <boost/test/included/unit_test_framework.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include "../Panorama/Panorama.h"

#include "../Panorama/Process.h"


using namespace boost::unit_test_framework;

using namespace PT;
using namespace std;

void Test_Lens()
{
    Lens l;
    bool b=l.readEXIF("pics/test1.jpg");
    BOOST_CHECK(b);
    // test lens;
    BOOST_CHECK_CLOSE(l.HFOV, 51.64, 0.01);
    BOOST_CHECK_CLOSE(l.focalLength, 5.41, 0.1);
    BOOST_CHECK_CLOSE(l.focalLengthConversionFactor, 6.88, 0.01);
}

void Test_PanoImage()
{

    Variable a("a"),b("a");
    BOOST_CHECK(a == b);
    a.link(0);
    BOOST_CHECK(! (a == b));
    b.link(1);
    BOOST_CHECK(! (a == b));
    b.link(0);
    BOOST_CHECK(a == b);
    b.link(1);
    a.unlink();
    b.unlink();
    BOOST_CHECK(a == b);

    PanoImage img("pics/test1.jpg");
    BOOST_CHECK_EQUAL(img.getFilename(), "pics/test1.jpg");
    ImageOptions iopt;
    BOOST_CHECK(img.getOptions() == iopt);
    iopt.featherWidth = 5;
    img.setOptions(iopt);
    BOOST_CHECK(img.getOptions() == iopt);
    BOOST_CHECK_EQUAL(img.getWidth(), (unsigned int) 1600);
    BOOST_CHECK_EQUAL(img.getHeight(), (unsigned int) 1200);
    BOOST_CHECK(img.getOptions().lensNr == 0);
    iopt = img.getOptions();
    iopt.lensNr = 1;
    img.setOptions(iopt);
    BOOST_CHECK(img.getOptions().lensNr == 1);
//    img.getVariables().print(std::cout) << std::endl;
//    var1.print(std::cout) << std::endl;

}

void Test_Panorama()
{
    Panorama p1;
    BOOST_CHECK_EQUAL(p1.getNrOfImages(), (unsigned int) 0);

    // add an image
    unsigned int nr = p1.addImage("pics/test1.jpg");
    BOOST_CHECK_EQUAL(nr, (unsigned int) 0);
    BOOST_CHECK_EQUAL(p1.getNrOfImages(), (unsigned int) 1);
    BOOST_CHECK_EQUAL(p1.getNrOfCtrlPoints(), (unsigned int) 0);
    BOOST_CHECK_EQUAL(p1.getImage(nr).getFilename(), "pics/test1.jpg");

    // add a control point in this image.
    ControlPoint cp1(0, 0, 0, 0, 10,500, ControlPoint::X);
    unsigned int cp = p1.addCtrlPoint(cp1);
    BOOST_CHECK_EQUAL(p1.getNrOfCtrlPoints(), (unsigned int) 1);
    BOOST_CHECK_EQUAL(cp,  (unsigned int) 0);
    BOOST_CHECK_EQUAL(p1.getImage(0).getOptions().lensNr, (unsigned int) 0);

    nr = p1.addImage("pics/test2.jpg");
    BOOST_CHECK_EQUAL(nr, (unsigned int) 1);
    BOOST_CHECK_EQUAL(p1.getNrOfImages(), (unsigned int) 2);
    BOOST_CHECK_EQUAL(p1.getImage(1).getFilename(), "pics/test2.jpg");

    // check get/set variables
    ImageVariables var1;
    var1.roll.setValue(1);
    var1.pitch.setValue(2);
    var1.yaw.setValue(3);
    var1.HFOV.setValue(4);
    var1.a.setValue(5);
    var1.b.setValue(6);
    var1.c.setValue(7);
    var1.d.setValue(8);
    var1.e.setValue(9);
    p1.setVariables(0,var1);

    VariablesVector realvar = p1.getVariables();
    VariablesVector targetvar;
    targetvar.resize(2);
    BOOST_CHECK(targetvar[0] == targetvar[1]);
    targetvar[1].a.link(0);
    targetvar[1].b.link(0);
    targetvar[1].c.link(0);
    targetvar[1].d.link(0);
    targetvar[1].e.link(0);
    targetvar[0].roll.setValue(1);
    targetvar[0].pitch.setValue(2);
    targetvar[0].yaw.setValue(3);
    targetvar[0].HFOV.setValue(4);
    targetvar[0].a.setValue(5);
    targetvar[0].b.setValue(6);
    targetvar[0].c.setValue(7);
    targetvar[0].d.setValue(8);
    targetvar[0].e.setValue(9);
    BOOST_CHECK(realvar == targetvar);
    targetvar[1].print(std::cout,false) << std::endl;
    realvar[1].print(std::cout, false) << std::endl;
    targetvar[1].print(std::cout, true) << std::endl;
    realvar[1].print(std::cout, true) << std::endl;

    // check single manipulation
    targetvar[0].roll.setValue(10);
    targetvar[1].roll.setValue(11);
    p1.setVariables(0,targetvar[0]);
    p1.setVariables(1,targetvar[1]);
    VariablesVector resvar = p1.getVariables();
    BOOST_CHECK(resvar == targetvar);

    // add another image
    nr = p1.addImage("pics/test3.jpg");
    BOOST_CHECK_EQUAL(nr, (unsigned int) 2);
    BOOST_CHECK_EQUAL(p1.getNrOfImages(), (unsigned int) 3);
    BOOST_CHECK_EQUAL(p1.getImage(2).getFilename(), "pics/test3.jpg");

    // add some control points
    cp = p1.addCtrlPoint(ControlPoint(0, 100,100, 1, 201, 201));
    BOOST_CHECK_EQUAL(cp,  (unsigned int) 1);
    cp = p1.addCtrlPoint(ControlPoint(0, 102,102, 1, 202, 202));
    BOOST_CHECK_EQUAL(cp,  (unsigned int) 2);
    cp = p1.addCtrlPoint(ControlPoint(1, 203,203, 2, 301, 301));
    BOOST_CHECK_EQUAL(cp,  (unsigned int) 3);
    cp = p1.addCtrlPoint(ControlPoint(1, 204,204, 2, 302, 302));
    BOOST_CHECK_EQUAL(cp,  (unsigned int) 4);

    BOOST_CHECK_EQUAL(p1.getNrOfCtrlPoints(), (unsigned int) 5);
    std::vector<unsigned int> points = p1.getCtrlPointsForImage(1);
    BOOST_CHECK_EQUAL(points.size(), (unsigned int)4);
    std::vector<unsigned int> pv;
    pv.push_back(1);
    pv.push_back(2);
    pv.push_back(3);
    pv.push_back(4);
    BOOST_CHECK(points == pv);

    ControlPoint cp1_2 = p1.getCtrlPoint(1);
    BOOST_CHECK(cp1_2 == ControlPoint(0, 100,100, 1, 201, 201));

    nr = p1.addImage("pics/test3.jpg");
    Lens l2;
    l2.HFOV = 10;
    p1.addLens(l2);
    p1.setLens(nr,1);

    OptimizeVector optSettings;
    optSettings.push_back(OptimizerSettings());
    optSettings.push_back(OptimizerSettings());
    optSettings.push_back(OptimizerSettings());
    optSettings.push_back(OptimizerSettings());
    optSettings.back().a = true;
    p1.printOptimizerScript(std::cout, optSettings);

    PanoramaOptions output;
    p1.printStitcherScript(std::cout, output);

    p1.removeImage(1);
    BOOST_CHECK_EQUAL(p1.getNrOfCtrlPoints(), (unsigned int) 1);
    BOOST_CHECK_EQUAL(p1.getNrOfImages(), (unsigned int) 3);

}

void Test_Process()
{
    Process proc;
    
    cerr << "sleep 10" << endl;
    proc.open("sleep 10");
    cerr << "executing..." << endl;
    proc.wait();
    cerr << "finished" << endl;
}
    
void Test_PanoCreation()
{
    Panorama pano;
    
    pano.addImage("pics/test1.jpg");
    pano.addImage("pics/test2.jpg");
    
    pano.addCtrlPoint(ControlPoint(0,1127,408, 1,210,377));
    pano.addCtrlPoint(ControlPoint(0,1303,585, 1,387,559));
    pano.addCtrlPoint(ControlPoint(0,1475,573, 1,551,546));
    pano.addCtrlPoint(ControlPoint(0,1479,434, 1,557,413));
    pano.addCtrlPoint(ControlPoint(0,988,405, 1,59,405));
    
    Process process;
    OptimizeVector optSettings;
    optSettings.push_back(OptimizerSettings());
    optSettings.push_back(OptimizerSettings());

    pano.runOptimizer(process, optSettings);
    VariablesVector vars = pano.getVariables();
    CPVector cps = pano.getCtrlPoints();

    process.wait();
    pano.readOptimizerOutput(vars, cps);
    
}

test_suite*
init_unit_test_suite( int, char** )
{
    test_suite* test= BOOST_TEST_SUITE( "Panorama tests" );
    test->add(BOOST_TEST_CASE(&Test_Lens));
    test->add(BOOST_TEST_CASE(&Test_PanoImage));
    test->add(BOOST_TEST_CASE(&Test_Panorama));
    test->add(BOOST_TEST_CASE(&Test_Process));
    test->add(BOOST_TEST_CASE(&Test_PanoCreation));
    return test;
}
